@@ -1,5 +1,6 @@
 """For Interfacing to Free-Wili Devices."""
 
+import datetime
 import pathlib
 import platform
 import sys
@@ -38,7 +39,7 @@ class FreeWili:
         self._display_serial: None | FreeWiliSerial = None
 
     def __str__(self) -> str:
-        return f"Free-Wili {self.device.serial}"
+        return f"{self.device.name} {self.device.serial} ({self.device.device_type.name})"
 
     def __repr__(self) -> str:
         return f"<{self.__class__.__name__}: {self.device.serial}>"
@@ -88,6 +89,19 @@ class FreeWili:
         -------
             None
         """
+        if (self.device.device_type == fwf.DeviceType.UF2) and (processor_type == FreeWiliProcessorType.Main):
+            # Special case for the DefCon 2024/2025 badge, where the main processor is in UF2 mode.
+            devs = self.device.get_usb_devices(fwf.USBDeviceType.MassStorage)
+            if devs:
+                return devs[0]
+        if self.device.device_type in (fwf.DeviceType.DefCon2024Badge, fwf.DeviceType.DefCon2025FwBadge):
+            if processor_type == FreeWiliProcessorType.Main:
+                # Special case for the DefCon 2024/2025 badge, where the main processor is in UF2 mode.
+                devs = self.device.get_usb_devices(fwf.USBDeviceType.SerialMain)
+                if devs:
+                    return devs[0]
+            else:
+                return None
         match processor_type:
             case FreeWiliProcessorType.Main:
                 devs = self.device.get_usb_devices(fwf.USBDeviceType.SerialMain)
@@ -166,8 +180,11 @@ class FreeWili:
             None | FreeWiliSerial:
                 FreeWiliSerial on success, None otherwise.
         """
+        is_badge: bool = False
+        if self.main and self.device.device_type in (fwf.DeviceType.DefCon2024Badge, fwf.DeviceType.DefCon2025FwBadge):
+            is_badge = True
         if not self._main_serial and self.main and self.main.port:
-            self._main_serial = FreeWiliSerial(self.main.port, self._stay_open, "Main: " + str(self))
+            self._main_serial = FreeWiliSerial(self.main.port, self._stay_open, "Main: " + str(self), is_badge)
         if self._main_serial:
             self._main_serial.stay_open = self._stay_open
         return self._main_serial
@@ -204,6 +221,14 @@ class FreeWili:
             Result[FreeWiliSerial, str]:
                 Ok(FreeWiliSerial) on success, Err(str) otherwise.
         """
+        if self.device.device_type in (fwf.DeviceType.DefCon2024Badge, fwf.DeviceType.DefCon2025FwBadge):
+            if processor_type in (FreeWiliProcessorType.Main, FreeWiliProcessorType.Display):
+                if self.main_serial:
+                    return Ok(self.main_serial)
+                else:
+                    return Err("Main serial isn't valid")
+            else:
+                return Err(f"{processor_type} isn't a valid FreeWiliSerial type")
         match processor_type:
             case FreeWiliProcessorType.Main:
                 if self.main_serial:
@@ -249,11 +274,11 @@ class FreeWili:
         if self.main_serial:
             result = self.main_serial.open(block, timeout_sec)
             if result.is_err():
-                return Err(result.err())
+                return Err(str(result.err()))
         if self.display_serial:
             result = self.display_serial.open(block, timeout_sec)
             if result.is_err():
-                return Err(result.err())
+                return Err(str(result.err()))
         return Ok(None)
 
     def close(self, restore_menu: bool = True) -> None:
@@ -434,6 +459,32 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
+    def load_fpga_from_file(
+        self, file_name: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Load FPGA from file.
+
+        Arguments:
+        ----------
+            file_name: str
+                Name of the file in the FreeWili. 8.3 filename limit exists as of V12
+                'spi','uart','i2c' or a filename name with extension.
+            processor: FreeWiliProcessorType
+                Processor to upload the file to.
+
+        Returns:
+        ---------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.load_fpga_from_file(file_name)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
     def get_io(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main) -> Result[tuple[int, ...], str]:
         """Get all the IO values.
 
@@ -469,8 +520,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -505,8 +556,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -523,7 +574,7 @@ class FreeWili:
         green: int,
         blue: int,
         processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
-    ) -> Result[ResponseFrame, str]:
+    ) -> Result[str, str]:
         """Set the GUI RGB LEDs.
 
         Parameters:
@@ -541,8 +592,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -552,6 +603,32 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
+    def read_write_spi_data(
+        self, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Read and Write SPI data.
+
+        Parameters:
+        -----------
+            data : bytes
+                The data to write.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+                The str is the response from the device, typically "OK".
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.read_write_spi_data(data)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
     def read_i2c(
         self, address: int, register: int, data_size: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
     ) -> Result[bytes, str]:
@@ -570,8 +647,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[bytes, str]:
+                Ok(bytes) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -634,7 +711,7 @@ class FreeWili:
 
     def show_gui_image(
         self, fwi_path: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
-    ) -> Result[ResponseFrame, str]:
+    ) -> Result[str, str]:
         """Show a fwi image on the display.
 
         Arguments:
@@ -646,8 +723,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -659,7 +736,7 @@ class FreeWili:
 
     def show_text_display(
         self, text: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
-    ) -> Result[ResponseFrame, str]:
+    ) -> Result[str, str]:
         """Show text on the display.
 
         Arguments:
@@ -671,8 +748,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -705,9 +782,7 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
-    def reset_display(
-        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
-    ) -> Result[ResponseFrame, str]:
+    def reset_display(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display) -> Result[str, str]:
         """Reset the display back to the main menu.
 
         Arguments:
@@ -717,8 +792,8 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
-                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
             case Ok(serial):
@@ -747,7 +822,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -777,7 +852,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -807,7 +882,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -834,7 +909,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -861,7 +936,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -888,7 +963,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -915,7 +990,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -942,7 +1017,7 @@ class FreeWili:
 
         Returns:
         ---------
-            Result[ResponseFrame, str]:
+            Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
         match self.get_serial_from(processor):
@@ -1458,6 +1533,132 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
+    def get_rtc(
+        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[tuple[datetime.datetime, int], str]:
+        """Get the RTC (Real-Time Clock) from the FreeWili.
+
+        Arguments:
+        ----------
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[tuple[datetime.datetime, int], str]:
+                Ok(tuple[datetime.datetime, int]) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.get_rtc()
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_rtc(
+        self,
+        dt: datetime.datetime,
+        trim: int | None = None,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
+    ) -> Result[str, str]:
+        """Set the RTC (Real-Time Clock) on the FreeWili.
+
+        Arguments:
+        ----------
+            dt: datetime
+                The datetime to set the RTC to.
+            trim: int | None
+                The trim value to set. (-127 - 127). If None, no trim is set.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_rtc(dt, trim)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_settings_to_default(self, processor: FreeWiliProcessorType) -> Result[str, str]:
+        """Set the settings to default on the FreeWili.
+
+        Notes: Settings are unstable as of v54 firmware. Subject to change in future firmware versions.
+
+        Arguments:
+        ----------
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_settings_to_default()
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_settings_as_startup(self, processor: FreeWiliProcessorType) -> Result[str, str]:
+        """Set the settings as startup on the FreeWili.
+
+        Notes: Settings are unstable as of v54 firmware. Subject to change in future firmware versions.
+
+        Arguments:
+        ----------
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_settings_as_startup()
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_system_sounds(
+        self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
+    ) -> Result[str, str]:
+        """Set the system sounds on the FreeWili.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable system sounds.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_system_sounds(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
 
 @dataclass(frozen=True)
 class FileMap:
